#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <inttypes.h>
#include <time.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#if defined(__NetBSD__) || defined(__OpenBSD__) || defined(__FreeBSD__)
# include <sys/dvdio.h>
#elif defined(__linux__)
# include <linux/cdrom.h>
#else
# error "Need the DVD ioctls"
#endif

#include "css_auth.h"
#include "css.h"

#ifdef __FreeBSD__
typedef struct dvd_struct   dvd_struct;
typedef struct dvd_authinfo dvd_authinfo;
#endif

typedef struct {
	uint8_t Challenge[10];
	uint8_t Key1[LEN_KEY];
	uint8_t Key2[LEN_KEY];
	uint8_t KeyCheck[LEN_KEY];
	uint8_t variant;
} DiscKeys;

DiscKeys discKeys;

#ifdef DEBUG
/**
 * Print Functions
 **/

static void _CSSPrintBytes (const uint8_t *val, int len)
{
	const uint8_t *ptr = val;

	for (; len > 0; len--)
		fprintf (stderr, " %02X", *ptr++ & 0xff);
}

#define _CSSPrintDebug(args...) fprintf (stderr, args)
#define _CSSPrintDebugBytes(descr, val, len) fprintf (stderr, descr":\t"); _CSSPrintBytes (val, len); fprintf (stderr, "\n")
#else
#define _CSSPrintDebug(args...)
#define _CSSPrintDebugBytes(descr, val, len)
#define _CSSPrintBytes(val, len)
#endif

/**
 * CSSAuthDrive
 * Determine if drive is authentic
 * Return: -1 if failed, 0..32 = variant of CSS
 **/


static int _CSSAuthDrive (const uint8_t *key, DiscKeys *disc)
{
	int i;

	for (i=0; i<LEN_KEY; i++)
		disc->Key1[i] = key[4-i];

	for (i=0; i<32; ++i) {
		CryptKey1( i, disc->Challenge, disc->KeyCheck);

		if (memcmp (disc->KeyCheck, disc->Key1, LEN_KEY) == 0) {
			_CSSPrintDebug ("Drive Authentic - using variant %d\n", i);
			disc->variant = i;
			return i;
		}
	}

	_CSSPrintDebug ("Drive would not Authenticate\n");

	return -1;
}


/**
 * Get Disc Key
 * Input:
 * fd 	device handle of DVD
 * agid
 * key 	player key
 * 
 * Output:
 * disckey	2048 byte block of disc key XOR player key
 * return 	0=failed, 1 = success
 **/

static int _CSSGetDiscKey (int fd, int agid, char *key, char *key_disc)
{
	dvd_struct s;
	int i;

#ifndef __FreeBSD__
	s.type = DVD_STRUCT_DISCKEY;
	s.disckey.agid = agid;
	memset (s.disckey.value, 0, 2048);

	if (ioctl(fd, DVD_READ_STRUCT, &s)<0) {
		perror ("Could not read Disc Key");
		return -1;
	}

	for (i=0; i<sizeof s.disckey.value; i++)
		s.disckey.value[i] ^= key[4 - (i % LEN_KEY)];

	_CSSPrintDebugBytes ("Received Disc Key", s.disckey.value, 10);

	memcpy (key_disc, s.disckey.value, 2048);
#else

    /*
     * FreeBSD's dvd_struct is not a union as it is with the other
     * platforms; hence, there are no additional sub-member
     * dereferences.
     */

    s.format = DVD_STRUCT_DISCKEY;
    s.agid   = agid;
    memset( s.data, 0, 2048);

    if (ioctl( fd, DVDIOCREADSTRUCTURE, &s) < 0) {
        perror( "Could not read Disc Key");
        return -1;
    }

    for( i=0; i<sizeof(s.data); i++)
        s.data[i] ^= key[4 - (i % LEN_KEY)];

	_CSSPrintDebugBytes( "Received Disc Key", s.data, 10);

	memcpy( key_disc, s.data, 2048);
#endif

	return 0;
}


/**
 * Get Title Key
 * Input:
 * fd 		device handle of DVD
 * agid
 * key 		player key
 * 
 * Output:
 * key_title	2048 byte block of disc key XOR player key
 * return 		0=failed, 1 = success
 **/

static int _get_title_key (int fd, int agid, int lba, char *key, char *key_title)
{
	dvd_authinfo ai;
	int i;

#ifndef __FreeBSD__
	ai.type = DVD_LU_SEND_TITLE_KEY;

	ai.lstk.agid = agid;
	ai.lstk.lba = lba;

	if (ioctl (fd, DVD_AUTH, &ai)) {
		perror ("GetTitleKey failed");
		return -1;
	}

	for (i=0; i< LEN_KEY; ++i)
		ai.lstk.title_key[i] ^= key[4 - (i%LEN_KEY)];

	_CSSPrintDebugBytes ("Received Title Key", ai.lstk.title_key, LEN_KEY);
	_CSSPrintDebug (" CPM=%d, CP_SEC=%d, CGMS=%d\n", ai.lstk.cpm, ai.lstk.cp_sec, ai.lstk.cgms);

	memcpy (key_title, ai.lstk.title_key, LEN_KEY);
#else
    ai.format = DVD_REPORT_TITLE_KEY;
    ai.agid   = agid;
    ai.lba    = lba;
    memset( ai.keychal, 0, 10);

    if (ioctl( fd, DVDIOCREPORTKEY, &ai)) {
        perror( "GetTitleKey failed");
        return -1;
    }

    for( i=0; i<LEN_KEY; i++)
        ai.keychal[i] ^= key[4 - (i%LEN_KEY)];

	_CSSPrintDebugBytes( "Received Title Key", ai.keychal, LEN_KEY);
	_CSSPrintDebug(" CPM=%d, CP_SEC=%d, CGMS=%d\n", ai.cpm, ai.cp_sec, ai.cgms);

	memcpy( key_title, ai.keychal, LEN_KEY);
#endif

	return 0;
}


/**
 * Get ASF (Authentication Success Flag)
 * fd 	device handle of DVD
 * Output:
 * return: 0 = failure, 1 = success
 **/

int CSSGetASF (int fd)
{
	dvd_authinfo ai;

#ifndef __FreeBSD__
	ai.type = DVD_LU_SEND_ASF;
        ai.lsasf.agid = 0;
	ai.lsasf.asf = 0;

        if (ioctl(fd, DVD_AUTH, &ai)) {
                perror ("GetASF");
                return -1;
        }

        _CSSPrintDebug ("%sAuthenticated", (ai.lsasf.asf) ? "" : "not ");

        return 0;

#else
	ai.format = DVD_REPORT_ASF;
	ai.agid   = 0;
	ai.asf    = 0;

	if (!(ioctl(fd, DVDIOCREPORTKEY, &ai))) {
		_CSSPrintDebug ("%sAuthenticated\n", (ai.asf) ? "" : "not");
		return 0;
	}
#endif
	perror ("GetASF");
	return -1;
}


/**
 *
 **/

int _CSSGetAgid (int fd, dvd_authinfo *ai)
{
	int tries;
	int rv;

	/* Init sequence, request AGID */
	for (tries = 1; tries < 4; ++tries) {
		_CSSPrintDebug ("Request AGID [%d]...", tries);
#ifndef __FreeBSD__
		ai->type = DVD_LU_SEND_AGID;
		ai->lsa.agid = 0;
		rv =  ioctl (fd, DVD_AUTH, ai);
#else
        ai->format = DVD_REPORT_AGID;
        ai->agid   = 0;
        rv         = ioctl( fd, DVDIOCREPORTKEY, ai);
#endif
		if (rv != -1) {
			// found key
            _CSSPrintDebug( "AGID %d\n", ai->agid);
			return( tries);
		}
        else {
            _CSSPrintDebug("\n");
        }
		perror ("N/A, invalidating");
#ifndef __FreeBSD__
		ai->type = DVD_INVALIDATE_AGID;
		ai->lsa.agid = 0;
		ioctl (fd, DVD_AUTH, ai);
#else
		ai->format = DVD_INVALIDATE_AGID;
		ai->agid   = 0;
		ioctl (fd, DVDIOCREPORTKEY, ai);
#endif
	}

	_CSSPrintDebug ("Cannot get AGID\n");

	return -1;
}


#ifndef __FreeBSD__

/**
 * Simulation of a non-CSS compliant host (i.e., the authentication fails,
 * but idea is here for a real CSS-compliant authentication scheme).
 **/

static int _CSSAuthHost (dvd_authinfo *ai, DiscKeys *disc)
{
        int i;

        switch (ai->type) {
        /* Host data receive (host changes state) */
        case DVD_LU_SEND_AGID:
                _CSSPrintDebug ("AGID %d", ai->lsa.agid);

                ai->type = DVD_HOST_SEND_CHALLENGE;
                break;

        case DVD_LU_SEND_KEY1:
                _CSSPrintDebugBytes ("LU sent key1", ai->lsk.key, LEN_KEY);

                if (_CSSAuthDrive (ai->lsk.key, disc)<0) {
                        ai->type = DVD_AUTH_FAILURE;
                        return -EINVAL;
                }
                ai->type = DVD_LU_SEND_CHALLENGE;
                break;

        case DVD_LU_SEND_CHALLENGE:
                for (i = 0; i < 10; ++i)
                        disc->Challenge[i] = ai->hsc.chal[9-i];

                _CSSPrintDebugBytes ("LU sent challenge", ai->hsc.chal, 10);

                CryptKey2 (disc->variant, disc->Challenge, disc->Key2);
                ai->type = DVD_HOST_SEND_KEY2;
                break;
        /* Host data send */
        case DVD_HOST_SEND_CHALLENGE:
                for (i = 0; i < 10; ++i)
                        ai->hsc.chal[9-i] = disc->Challenge[i];

                _CSSPrintDebugBytes ("Host sending challenge", ai->hsc.chal, 10)
;

                /* Returning data, let LU change state */
                break;

        case DVD_HOST_SEND_KEY2:
                for (i = 0; i < LEN_KEY; ++i)
                        ai->hsk.key[4-i] = disc->Key2[i];

                _CSSPrintDebugBytes ("Host sending key 2", &disc->Key2[0], LEN_K
EY);

                /* Returning data, let LU change state */
                break;

        default:
                _CSSPrintDebug ("Got invalid state %d", ai->type);

                return -EINVAL;
        }

        return 0;
}
#endif


/**
 *
 **/

int CSSAuth (int fd, DiscKeys *disc)
{
 	dvd_authinfo ai;
	int i, agid;

	memset (&ai, 0, sizeof (ai));

	if (CSSGetASF (fd))
		return -1;

	_CSSGetAgid (fd, &ai);

	for (i = 0; i < 10; ++i)
		disc->Challenge[i] = i;

#ifndef __FreeBSD__
	/* Send AGID to host */
	if (_CSSAuthHost(&ai, disc) < 0) {
		_CSSPrintDebug ("Send AGID to host failed\n");
		return -1;
	}
	/* Get challenge from host */
	if (_CSSAuthHost(&ai, disc) < 0) {
		_CSSPrintDebug ("Get challenge from host failed\n");
		return -1;
	}
	agid = ai.lsa.agid;
	/* Send challenge to LU */
	if (ioctl(fd, DVD_AUTH, &ai) < 0) {
		_CSSPrintDebug ("Send challenge to LU failed\n");
		return -1;
	}
	/* Get key1 from LU */
	if (ioctl(fd, DVD_AUTH, &ai) < 0) {
		_CSSPrintDebug ("Get key1 from LU failed\n");
		return -1;
	}
	/* Send key1 to host */
	if (_CSSAuthHost(&ai, disc) < 0) {
		_CSSPrintDebug ("Send key1 to host failed\n");
		return -1;
	}
	/* Get challenge from LU */
	if (ioctl(fd, DVD_AUTH, &ai) < 0) {
		_CSSPrintDebug ("Get challenge from LU failed\n");
		return -1;
	}
	/* Send challenge to host */
	if (_CSSAuthHost (&ai, disc) < 0) {
		_CSSPrintDebug ("Send challenge to host failed\n");
		return -1;
	}
	/* Get key2 from host */
	if (_CSSAuthHost (&ai, disc) < 0) {
		_CSSPrintDebug ("Get key2 from host failed\n");
		return -1;
	}
	/* Send key2 to LU */
	if (ioctl(fd, DVD_AUTH, &ai) < 0) {
		_CSSPrintDebug ("Send key2 to LU failed (expected)\n");
		return -1;
	}

	if (ai.type == DVD_AUTH_ESTABLISHED)
		_CSSPrintDebug ("DVD is authenticated\n");
	else if (ai.type == DVD_AUTH_FAILURE)
		_CSSPrintDebug ("DVD authentication failed\n");
#else
	agid = ai.agid;

	/* Get challenge from host */

	ai.format = DVD_SEND_CHALLENGE;
	for (i=0; i<10; ++i)
		ai.keychal[9-i] = disc->Challenge[i];

	_CSSPrintDebug("Host sending challenge: ");
    _CSSPrintBytes(ai.keychal, 10);
    _CSSPrintDebug("\n");

	/* Send challenge to LU */

	if (ioctl(fd, DVDIOCSENDKEY, &ai) < 0) {
		_CSSPrintDebug("Send challenge to LU failed\n");
		return -1;
	}

	/* Get key1 from LU */

	ai.format = DVD_REPORT_KEY1;
	if (ioctl(fd, DVDIOCREPORTKEY, &ai) < 0) {
		_CSSPrintDebug("Get key1 from LU failed\n");
		return -1;
	}

	/* Send key1 to host */

	_CSSPrintDebug("LU sent key1: ");
    _CSSPrintBytes(ai.keychal, LEN_KEY);
    _CSSPrintDebug("\n");

    if (_CSSAuthDrive( ai.keychal, disc) < 0) {
		_CSSPrintDebug("Send key1 to host failed\n");
		return -1;
    }

	/* Get challenge from LU */

	ai.format = DVD_REPORT_CHALLENGE;
	if (ioctl(fd, DVDIOCREPORTKEY, &ai) < 0) {
		_CSSPrintDebug("Get challenge from LU failed\n");
		return -1;
	}

	/* Send challenge to host */

	for (i=0; i<10; ++i)
		disc->Challenge[i] = ai.keychal[9-i];

	_CSSPrintDebug("LU sent challenge: ");
    _CSSPrintBytes(disc->Challenge, 10);
    _CSSPrintDebug("\n");

	CryptKey2(disc->variant, disc->Challenge, disc->Key2);

	/* Get key2 from host */

	ai.format = DVD_SEND_KEY2;
	for (i=0; i<LEN_KEY; ++i)
		ai.keychal[4-i] = disc->Key2[i];

	_CSSPrintDebug("Host sending key 2: ");
    _CSSPrintBytes(disc->Key2, LEN_KEY);
    _CSSPrintDebug("\n");

	/* Send key2 to LU */
	if (ioctl(fd, DVDIOCSENDKEY, &ai) < 0) {
		printf("Send key2 to LU failed (expected)\n");
		return -1;
	}

#endif

	memcpy (disc->Challenge, disc->Key1, LEN_KEY);
	memcpy (disc->Challenge+LEN_KEY, disc->Key2, LEN_KEY);
	CryptBusKey (disc->variant, disc->Challenge, disc->KeyCheck);

	_CSSPrintDebugBytes ("Received Session Key\n", &disc->KeyCheck[0], LEN_KEY);

	return( agid);
}


/**
 * CSSAuthDisc: Authorise disc
 * fd	device
 * key	player key
 **/
	
int CSSAuthDisc (int fd, char *key)
{
	int agid;
	DiscKeys *disc = &discKeys;

	agid = CSSAuth( fd, disc);

	if (agid < 0)
		return -1;

	if (CSSGetASF (fd) < 0)
		return -1;

	_CSSGetDiscKey (fd, agid, disc->KeyCheck, key);

	if (CSSGetASF (fd) < 0)
		return -1;

	return 0;
}


/** 
 * CSSAuthTitle
 * Authorise title from disc
 * fd	device
 * key	disc key
 * lba	logical block address of title	
 **/

int CSSAuthTitle (int fd, char *key, int lba)
{
	int agid;
	DiscKeys *disc = &discKeys;

	agid = CSSAuth( fd, disc);
	if (agid < 0) return -1;

	_get_title_key (fd, agid, lba, disc->KeyCheck, key);

	if (CSSGetASF (fd) < 0)
		return -1;

	return 0;
}


/**
 * CSSisEncrypted
 * check if disk is encrypted at all
 * fd	device
 **/

int CSSisEncrypted (int fd)
{
	dvd_struct dvd;

#ifndef __FreeBSD__
	dvd.copyright.type = DVD_STRUCT_COPYRIGHT;
	dvd.copyright.layer_num = 0;

	if (ioctl (fd, DVD_READ_STRUCT, &dvd) < 0) {
		_CSSPrintDebug ("error using ioctl DVD_READ_STRUCT\n");
		return -1;
	}

	return dvd.copyright.cpst;      // 0 ... not encrypted
#else
    dvd.format    = DVD_STRUCT_COPYRIGHT;
    dvd.layer_num = 0;

    if (ioctl( fd, DVDIOCREADSTRUCTURE, &dvd) < 0) {
		_CSSPrintDebug ("error using ioctl DVDIOCREADSTRUCTURE\n");
		return -1;
    }

    return dvd.cpst;                // 0 ... not encrypted
#endif
}
